/**
 * @author Brace Stout
 * 
 * This software is copyright 2006, 2007 by StoutWare Engineering
 * All rights are retained.
 * 
 * This is example code only.  No representation is made
 * regarding utility, correctness, merchantability, or
 * fitness for a particular purpose.
 *
 * You have permission to use this software and to make
 * derivative products.  In all cases, you must retain
 * this notice in derivative products as well as credit
 * the source(s).
 * 
 * You may not use this software or any derivative of this
 * software in a commercial product without the express
 * written consent of StoutWare Engineering.
 * 
 * You may not use this software or any derivative of this software
 * in any application where failure of this software to perform as
 * intended would jeopardize the safety of any individual or property.
 * 
 * StoutWare Engineering will in no way be responsible for
 * any damages whatsoever arising out of your use of this
 * or any derivation of this product.
 **/

/*
 * Implementation Notes:
 *
 * LoCoMoCo stands for Low Cost Motor Controller, and this
 * is the software driver for the hardware of the same name.
 * 
 * The serial-to-parallel converter requires that the least
 * significant bit _ALWAYS_ be one, so that it can perform
 * timing on the start bit to determine the required
 * sampling rate.  The next three bits are an address,
 * which is hardwired to 111 on LoCoMoCo.  Sending in
 * a spare byte with a low nibble of 0x01 synchronizes
 * the converter with the data stream.
 * 
 * When selecting the duty cycle, one should make sure that
 * there is enough free time in the cycle to support the
 * range of pulse widths needed. 
 *
 * A single update uses 22 bits.  There will normally be
 * two updates per motor on each duty cycle.  Assuming that
 * both motor controls have the same duty cycle, this runs
 * out to 88 bits per duty cycle.  It is important that
 * the duty cycle not be shorter than the time it takes for
 * 88 bits to be transmitted at the selected baud rate, or
 * performance will degrade.
 * 
 * For 2400 bps, this means a maximum of 27.3 cycles per second,
 * or one every 36.7 mS.  It is undesirable to set the update
 * rate so low, however, as the time of transmission would
 * overwhelm the duty cycling and provide a 50-50 duty cycle,
 * with very little room for variations of pulse width.
 */
package motor;

import java.io.IOException;
import java.io.OutputStream;

import javax.comm.CommPortIdentifier;
import javax.comm.SerialPort;

public class LoCoMoCo {

    public final static boolean ACTIVE_DEFAULT = false;
    public final static long DUTY_CYCLE_DEFAULT = 50;
    public final static int BAUD_RATE_DEFAULT = 2400;

    /*
     * The following parameters are required characteristics of the serial data
     * stream, due to the requirements of the serial-to-parallel converter chip
     * (SN74LV8153).
     */
    private static final int DATABITS = SerialPort.DATABITS_8;
    private static final int PARITY = SerialPort.PARITY_NONE;

    /*
     * Two stop bits are used to provide more reliable synchronization
     * characteristics.
     */
    private static final int STOPBITS = SerialPort.STOPBITS_2;

    /*
     * Communications is via the serial port. No input stream is needed - the
     * motor controller is an output device only.
     */
    private SerialPort m_port = null;
    private OutputStream m_outStream = null;

    /*
     * Two motor controls are implemented - one for the left motor and one for
     * the right motor.
     */
    private MotorControlImpl m_leftControl = null;
    private MotorControlImpl m_rightControl = null;

    /*
     * The updater manages sending updates to the device in a timely fashion, to
     * minimize latencies where possible.
     */
    private Updater m_updater = null;

    /**
     * @param active
     *            refers to whether the motor controller actively brakes during
     *            the inactive portion of the duty cycle. This may result in
     *            improved repeatability and control, especially while operating
     *            on inclines, but may also lead to higher current draw /
     *            shorter battery life. For motors with a high current draw,
     *            this may cause the L298 driver chip to generate significant
     *            amounts of heat.
     */
    private LoCoMoCo(SerialPort port, boolean active) throws IOException {

        m_port = port;
        m_updater = new Updater(m_port.getOutputStream(), 1.0 / port.getBaudRate());

        /*
         * The following values were used for the LoCoMoCo prototype
         * 
         * m_leftControl = new MotorControlImpl( active, 0x80, 0x40, 0x20 );
         * m_rightControl = new MotorControlImpl( active, 0x08, 0x04, 0x02 );
         */

        /*
         * These values correspond to LoCoMoCo v1.0 production board.
         */
        m_leftControl = new MotorControlImpl(active, 0x02, 0x01, 0x04);
        m_rightControl = new MotorControlImpl(active, 0x20, 0x10, 0x40);
    }

    /*
     * open() methods for establishing communications with LoCoMoCo
     * 
     * null value is returned if communications is not established
     */
    public static LoCoMoCo open(String comport) {
        return open(comport, BAUD_RATE_DEFAULT, ACTIVE_DEFAULT);
    }

    public static LoCoMoCo open(String comport, boolean active) {
        return open(comport, BAUD_RATE_DEFAULT, active);
    }

    private static LoCoMoCo open(String comport, int baudRate, boolean active) {
        try {
            CommPortIdentifier portId = CommPortIdentifier.getPortIdentifier(comport);
            SerialPort port = (SerialPort) portId.open("L2Bot Serial Motor Controller", 1000);
            port.setSerialPortParams(baudRate, DATABITS, STOPBITS, PARITY);
            return new LoCoMoCo(port, active);
        } catch (Throwable t) {
            return null;
        }
    }

    /*
     * Shut down LoCoMoCo communications.
     */
    public void close() {
        if (m_leftControl != null) {
            m_leftControl.shutdown();
            m_leftControl = null;
        }
        if (m_rightControl != null) {
            m_rightControl.shutdown();
            m_rightControl = null;
        }
        if (m_updater != null) {
            m_updater.shutdown();
            m_updater = null;
        }
        if (m_outStream != null) {
            m_outStream = null;
        }
        if (m_port != null) {
            m_port.close();
            m_port = null;
        }
    }

    public MotorControl getLeftMotorControl() {
        return m_leftControl;
    }

    public MotorControl getRightMotorControl() {
        return m_rightControl;
    }

    private class Updater implements Runnable {

        private boolean m_shutdown = false;

        @SuppressWarnings("unused")
        private boolean m_updated = false;

        @SuppressWarnings("unused")
        private boolean m_synchronized = false;

        private Thread m_thread = null;

        private double m_bitPeriod = 0;

        private int m_value = -1;

        public Updater(OutputStream outStream, double bitPeriod) {
            m_outStream = outStream;
            m_bitPeriod = bitPeriod;
            m_thread = new Thread(this);
            m_thread.setPriority(Thread.MAX_PRIORITY - 2);
            m_thread.start();
        }

        private Thread shutdownHook = new Thread(new Runnable() {
            public void run() {
                System.out.println("Shutdown hook start.");
                try {
                    synchronized (Updater.this) {
                        m_shutdown = true;
                        notifyAll();
                        m_thread.join();
                        m_outStream.write(0x01);
                        m_outStream.write(0x0f);
                        m_outStream.write(0x0f);
                        m_outStream.flush();
                    }
                } catch (Throwable t) {
                }
                System.out.println("Shutdown hook completion.");
                System.out.flush();
            }
        });

        public synchronized void run() {

            Runtime.getRuntime().addShutdownHook(shutdownHook);

            try {

                /*
                 * Send an initial synchronization byte to the device, and set
                 * the pending time to be 11 bit times.
                 */
                synchronized (m_outStream) {
                    m_outStream.write(0x01);
                    m_outStream.flush();
                }
                m_synchronized = true;
                double pending = 11 * m_bitPeriod;

                /* Loop until shutdown is requested. */
                while (!m_shutdown) {

                    /*
                     * Wait for the remaining transmission time, if any.
                     * 
                     * During this wait, m_value can be modified as needed. When
                     * modified, the m_updated flag will be set.
                     */
                    long delay = (long) (1000 * pending);
                    if (delay == 0) {
                        delay = DUTY_CYCLE_DEFAULT;
                    }

                    long reference = System.currentTimeMillis();
                    wait(delay);

                    /*
                     * Determine how much transmission time remains.
                     */
                    double elapsed = 0.001 * (System.currentTimeMillis() - reference);
                    pending = Math.max(0, pending - elapsed);

                    /*
                     * Hold off on sending additional data until output buffer
                     * is (nearly) empty.
                     */
                    if (pending <= 0.001) {

                        /*
                         * Write data and a synchronization byte to the command
                         * stream.
                         */
                        // System.out.printf( "Writing: %02x\n", m_value );
                        synchronized (m_outStream) {
                            m_outStream.write(0x0f | ((m_value << 4) & 0xf0));
                            m_outStream.write(0x0f | (m_value & 0xf0));
                            m_outStream.write(0x01);
                            m_outStream.flush();
                        }

                        pending = 34 * m_bitPeriod;
                        m_synchronized = true;

//                         if ( m_updated ) {
//                        
//                         /*
//                         * Value has changed - write corresponding bytes to
//                         the output
//                         * stream and clear the updated flag. Adjust the
//                         remaining
//                         * period.
//                         */
//                         m_updated = false;
//                        							
//                         m_outStream.write( 0x0f | ( ( m_value << 4 ) & 0xf0 )
//                         );
//                         m_outStream.write( 0x0f | ( m_value & 0xf0 ) );
//                         m_outStream.flush();
//                        
//                         /* Add 22 bit times to the remaining period */
//                         pending += 22 * m_bitPeriod;
//                         m_synchronized = false;
//                        
//                         } else if ( !m_synchronized ) {
//                        
//                         /* Write a synchronization byte to the command
//                         stream. */
//                         m_outStream.write( 0x01 );
//                         m_outStream.write( 0x0f | ( ( m_value << 4 ) & 0xf0 )
//                         );
//                         m_outStream.write( 0x0f | ( m_value & 0xf0 ) );
//                         m_outStream.flush();
//                        
//                         pending = 33 * m_bitPeriod;
//                         m_synchronized = true;
//                        
//                         } else {
//                        
//                         /* Write a synchronization byte to the command
//                         stream. */
//                         m_outStream.write( 0x01 );
//                         m_outStream.write( 0x0f | ( ( m_value << 4 ) & 0xf0 )
//                         );
//                         m_outStream.write( 0x0f | ( m_value & 0xf0 ) );
//                         m_outStream.flush();
//                        
//                         pending = 33 * m_bitPeriod;
//                         m_synchronized = true;
//                        
//                         }

                    }

                }

            } catch (Throwable t) {
            }

            Runtime.getRuntime().removeShutdownHook(shutdownHook);
            shutdownHook = null;

        }

        public synchronized void update(int mask, int value) {
            int newValue = 0xff & ((m_value & ~mask) | (value & mask));
            if (newValue != m_value) {
                m_value = newValue;
                m_updated = true;
                notifyAll();
            }
        }

        public void shutdown() {
            synchronized (this) {
                m_shutdown = true;
                notifyAll();
            }
            try {
                m_thread.join();
            } catch (InterruptedException e) {
            }
            m_thread = null;
        }

    }

    private class MotorControlImpl implements MotorControl, Runnable {

        private int m_forwardMask = 0;
        private int m_reverseMask = 0;
        private int m_enableMask = 0;
        private int m_mask = 0;
        private boolean m_brake = true;
        private boolean m_reverse = false;
        private boolean m_pushPull = false;
        private long m_dutyCycle = DUTY_CYCLE_DEFAULT;
        private long m_activeDuration = 0;
        private long m_inactiveDuration = DUTY_CYCLE_DEFAULT;
        private boolean m_shutdown = false;
        private Thread m_thread = null;

        public MotorControlImpl(boolean pushPull, int forwardMask, int reverseMask, int enableMask) {
            m_forwardMask = forwardMask;
            m_reverseMask = reverseMask;
            m_enableMask = enableMask;
            m_pushPull = pushPull;
            m_mask = forwardMask | reverseMask | enableMask;
            m_dutyCycle = DUTY_CYCLE_DEFAULT;
            m_thread = new Thread(this);
            m_thread.setPriority(Thread.MAX_PRIORITY - 1);
            m_thread.start();
        }

        public synchronized void setPower(double ratio) {
            m_brake = false;
            m_reverse = ratio < 0;
            double magnitude = Math.min(1, Math.abs(ratio));
            m_activeDuration = (long) (magnitude * (double) m_dutyCycle);
            m_inactiveDuration = m_dutyCycle - m_activeDuration;
        }

        public synchronized void setBrake() {
            m_brake = true;
            m_activeDuration = 0;
            m_inactiveDuration = m_dutyCycle;
        }

        public synchronized void setDutyCycle(long interval) {
            m_dutyCycle = interval;
        }

        public void shutdown() {
            synchronized (this) {
                m_shutdown = true;
                notifyAll();
            }
            try {
                m_thread.join();
            } catch (InterruptedException e) {
            }
            m_thread = null;
        }

        public synchronized void run() {

            try {

                m_updater.update(m_mask, m_pushPull ? m_enableMask : 0x00);

                boolean active = false;

                while (!m_shutdown) {

                    if (m_brake) {

                        active = false;
                        m_updater.update(m_mask, m_mask);
                        wait(m_dutyCycle);

                    } else {

                        if (active) {
                            active = (m_inactiveDuration == 0);
                        } else {
                            active = (m_activeDuration > 0);
                        }

                        if (active) {

                            m_updater.update(m_mask, m_enableMask | (m_reverse ? m_reverseMask : m_forwardMask));
                            wait(m_activeDuration);

                        } else {

                            /*
                             * Motor will either coast or be actively braked
                             * during the inactive portion of the duty cycle.
                             */
                            m_updater.update(m_mask, m_pushPull ? m_enableMask : m_forwardMask | m_reverseMask);
                            wait(m_inactiveDuration);

                        }

                    }

                }

                m_updater.update(m_mask, m_pushPull ? m_enableMask : 0x00);

            } catch (InterruptedException e) {
            }

        }

    }

}
